const fs = require('fs')
const os = require('os')
const md5 = require('md5')
const { promisify } = require('util')
const PythonWorker = require('./pythonWorker')

const readdir = promisify(fs.readdir)
const stat = promisify(fs.stat)

/**
 * INTERFACE BACKEND
 *
 * Communication with workers (for now only Python)
 * The workers are implemented as REPL clients that
 * communicate with JSON.
 *
 * Names:
 *  * Worker: a Python file that implements a REPL client (e.g. serial)
 *  * Port:
 *  * Connection:
 */

const WORKER_DIR = `${__dirname}/python_workers`
const HOST = os.hostname()

const state = {
  workers: [],
  ports: [],
  connections: [],
  lastScan: {
    workers: Date.now() - 100000,
    ports: Date.now() - 100000,
    connections: Date.now() - 100000
  }
}

const typeDefs = `
  type Worker {
    id: ID!
    interfaceName: String!
    workerScript: String!
    mtime: DateTime!
    updated: DateTime
    workerProcess: WorkerProcess
    ports: [Port]!
    connections: [Connection]!
    options: [Option!]
  }

  type WorkerProcess {
    pid: Int
    killed: Boolean!
    signalCode: String
    exitCode: Int
    spawnfile: String!
    spawnargs: [String]!
    error: [String]!
    data: [String]!
  }

  type Option {
    name: String!
    type: String!
    description: String
    values: [String!]
  }

  type Port {
    id: String!
    interfaceName: String!
    host: String!
    device: String!
    name: String
    description: String
  }

  type Connection {
    id: ID!
    interfaceName: String!
    host: String!
    device: String!
    workerProcess: WorkerProcess
  }

  extend type Query {
    workers: [Worker]!
    worker(id: ID, interfaceName: String): Worker!
    ports(interfaceName: String, force: Boolean): [Port]!
    connections: [Connection]!
    connection(id: ID!): Connection!
  }

  extend type Mutation {
    connect(interfaceName: String!, device: String!): Connection!
    spawnWorker(id: ID!): Worker!
    endWorker(id: ID!): Worker!
    killWorker(id: ID!): Worker!
    connectionCommand(connectionId: ID!, type: String!, string: String!, options: String): String!
  }
`

/**
 * WORKER SECTION
 */
async function findWorkers() {
  // 1. Don't check more frequently than once per second.
  // if (state.lastScan.workers + 1000 > Date.now()) return null
  // state.lastScan.workers = Date.now()

  // 2. Find all files in ./python_workers that end in _worker.py
  const fileNames = await readdir(WORKER_DIR)
  const workerFiles = fileNames.filter(fileName => fileName.includes('_worker.py'))

  // 3. For every worker script
  const workerPromises = workerFiles.map(async workerFile => {
    const interfaceName = workerFile.replace(/_worker\.py/, '')
    const workerScript = `${WORKER_DIR}/${workerFile}`
    // a. Find out if it was modified.
    const { mtime } = await stat(workerScript)
    const foundWorker = state.workers.find(worker => worker.interfaceName === interfaceName)
    if (foundWorker) {
      // b. If it was modified, save the modification time.
      if (foundWorker.mtime < mtime) foundWorker.updated = mtime
      return
    }
    // c. Spawn a new worker connection.
    const workerProcess = new PythonWorker(workerScript)
    const { data, error, pythonError } = await workerProcess.spawn()
    if (error) throw new Error(error)
    if (pythonError) throw new Error(pythonError)
    // d. Save the worker in the state.
    state.workers.push({
      id: md5(`${interfaceName}${workerScript}${mtime}`),
      interfaceName,
      workerScript,
      mtime,
      updated: null,
      workerProcess,
      ports: [],
      connections: [],
      options: []
    })
  })
  await Promise.all(workerPromises)
}

async function workers(parent, args, context, info) {
  await findWorkers()
  return state.workers.map(worker => ({
    ...worker,
    // Find ports
    ports: ports(worker.interfaceName, args, context, info),
    // Serialize worker process
    workerProcess: workerProcess(worker.workerProcess, args, context, info)
  }))
}

async function worker(parent, args, context, info) {
  await findWorkers()
  const { id, interfaceName } = args
  if (!id && !interfaceName) throw new Error('Either id or interfaceName needs to be provided!')
  const worker = state.workers.find(worker => worker.id === id || worker.interfaceName === interfaceName)
  if (!worker) throw new Error(`Worker id=${id}, interfaceName=${interfaceName} not found`)
  return {
    ...worker,
    workerProcess: workerProcess(worker.workerProcess, args, context, info)
  }
}

/**
 * PORTS SECTION
 */
async function findPorts() {
  console.log('find ports.')
  // 1. Make sure, workers are updated.
  await findWorkers()

  // 2. Don't check more frequently than once per second.
  // if (state.lastScan.ports + 1000 > Date.now()) return null
  // state.lastScan.ports = Date.now()

  // 3. Loop through all workers to find available ports.
  const portsPromises = state.workers.map(async worker => {
    // a) Ask worker for ports.
    const { data, error, pythonError } = await worker.workerProcess.send({ type: 'ports' })
    if (error) throw new Error(error)
    if (pythonError) throw new Error(pythonError)
    console.log(data)
    // b) Add all ports that are not in the list.
    data.forEach(port => {
      const id = port.name || port.device
      if (state.ports.find(port => port.id === id)) return null
      const newPort = {
        id,
        interfaceName: worker.interfaceName,
        host: os.hostname(),
        ...port
      }
      state.ports.push(newPort)
      worker.ports.push(newPort)
    })
  })
  await Promise.all(portsPromises)

  console.log('found ports.')
}

async function ports(parent, args, ctx, info) {
  await findPorts()
  const { interfaceName } = args
  const ifName = interfaceName || parent

  console.log(ifName)
  if (ifName) {
    const iface = state.workers.find(iface => iface.interfaceName === ifName)
    if (!iface) throw new Error(`Interface ${ifName} not found.`)
    return iface.ports
  } else {
    return state.ports
  }
}

async function port(parent, args, ctx, info) {
  await findPorts()
  const { id } = args
  if (!id) throw new Error('Need an id.')
  const port = state.ports.find(port => port.id === id)
  return port
}

async function findOptions() {
  // 1. Make sure, workers are updated.
  await findWorkers()

  // 2. Don't check more frequently than once per second.
  // if (state.lastScan.options + 1000 > Date.now()) return null
  // state.lastScan.options = Date.now()

  const optionPromises = state.workers.map(async worker => {
    const { data, error, pythonError } = await worker.workerProcess.pythonShell.send({ type: 'options' })
    if (error) throw new Error(error)
    if (pythonError) throw new Error(pythonError)
    iface.options.push(...data)
  })
  await Promise.all(optionPromises)
}

async function options(parent, args, ctx, info) {
  if (!parent) throw new Error('Need a parent.')
  const iface = state.interfaces.find(iface => iface.interfaceName === parent)

  // Try to find options if necessary
  if (!iface.options.length) await findOptions(parent)
  return iface.options
}

function workerProcess(parent, args, ctx, info) {
  let worker
  const { id } = args
  if (!id && !parent) throw new Error('Need either id or parent of worker.')
  if (parent) {
    worker = parent
  } else {
    worker = state.workers.find(worker => worker.id === id).workerProcess
  }
  const { killed, exitCode, signalCode, spawnargs, spawnfile, pid } = parent.pythonShell
  const { error, data } = parent
  return {
    pid,
    killed,
    exitCode,
    signalCode,
    spawnfile,
    spawnargs,
    error,
    data
  }
}

async function connections(parent, args, ctx, info) {
  if (parent) {
    const iface = state.interfaces.find(iface => iface.interfaceName === parent)
    return iface.connections
  } else {
    return state.connections
  }
}

async function connection(parent, args, context, info) {
  const connection = state.connections.find(connection => connection.id === args.id)
  return connection
}

async function connect(parent, args, ctx, info) {
  const { interfaceName, device } = args
  const iface = state.interfaces.find(iface => iface.interfaceName === interfaceName)
  const id = md5(interfaceName + device)
  if (iface.connections.find(connection => connection.id === id)) throw new Error('already connected.')
  const pythonWorker = new PythonWorker(iface.workerScript)
  const spawnData = await pythonWorker.spawn()
  if (spawnData.error) throw new Error(spawnData.error)
  const connection = {
    id,
    device,
    interfaceName,
    host: os.hostname(),
    worker: pythonWorker,
    workerProcess: (parent, args, context, info) => workerProcess(pythonWorker, args, context, info)
  }
  const connectionData = await connection.worker.send({ type: 'connect', device })
  if (connectionData.error) throw new Error(connectionData.error)
  if (connectionData.pythonError) throw new Error(connectionData.pythonError)
  iface.connections.push(connection)
  state.connections.push(connection)
  return connection
}

async function connectionCommand(parent, args, ctx, info) {
  const { connectionId, type, string, options } = args
  const connection = state.connections.find(connection => connection.id === connectionId)
  const { data, error, pythonError } = await connection.worker.send({ type, string, options })
  if (error) throw new Error(JSON.stringify(error))
  if (pythonError) throw new Error(JSON.stringify(pythonError))
  return data.response
}

// TODO Also find connections in interfaces.
async function endWorker(parent, args, ctx, info) {
  const { id } = args
  const connection = state.connections.find(connection => connection.id === id)
  const { data, error, pythonError } = await connection.worker.end()
  if (error) throw new Error(JSON.stringify(error))
  if (pythonError.error !== 0) throw new Error(JSON.stringify(pythonError))
  return connection.workerProcess()
}

async function spawnWorker(parent, args, ctx, info) {
  const { id } = args
  const connection = state.connections.find(connection => connection.id === id)
  const { data, error, pythonError } = await connection.worker.spawn()
  console.log(data, error, pythonError)
  if (error) throw new Error(JSON.stringify(error))
  if (pythonError) throw new Error(JSON.stringify(pythonError))
  const connectionData = await connection.worker.send({ type: 'connect', device: connection.device })
  if (connectionData.error) throw new Error(connectionData.error)
  if (connectionData.pythonError) throw new Error(connectionData.pythonError)
  return connection.workerProcess()
}

async function killWorker(parent, args, ctx, info) {
  const { id } = args
  const connection = state.connections.find(connection => connection.id === id)
  const { data, error, pythonError } = await connection.worker.kill()
  console.log(data, error, pythonError)
  if (error) throw new Error(JSON.stringify(error))
  if (pythonError) throw new Error(JSON.stringify(pythonError))
  return connection.workerProcess()
}

const resolvers = {
  Query: {
    workers,
    worker,
    ports,
    connections,
    connection
  },
  Mutation: {
    connect,
    connectionCommand,
    killWorker,
    endWorker,
    spawnWorker
  }
}

module.exports = { typeDefs, resolvers }
