const { spawn } = require('child_process')
const { Lock } = require('semaphore-async-await')
const { serializeError } = require('./utils')

function PythonWorker (workerScript, shellOptions) {
  // create a lock for incomming commands
  this.commandLock = new Lock()
  // create a lock for python communication
  this.pythonLock = new Lock()
  this.data = []
  this.error = []
  this.workerScript = workerScript
  // The python shell is started with this.spawn
  this.pythonShell = null

  this.transaction = async (func, args) => {
    console.log('transaction.')
    // 1. Block new commands
    await this.commandLock.acquire()
    // 2. Send data to Python shell (lock will be released by data/error event)
    await this.pythonLock.acquire()
    console.log('function call')
    func(args)
    console.log('function ended', this.pythonLock)
    // 3. Wait for data from the Python shell
    await this.pythonLock.acquire()
    const pythonError = this.error.pop()
    const workerResult = this.data.pop()
    this.pythonLock.release()
    console.log('results', { pythonError, ...workerResult })
    // 4. Unblock new commands
    this.commandLock.release()
    // 5. Return result
    console.log('transaction done.', { ...workerResult, pythonError })
    return { ...workerResult, pythonError }
  }

  this.isProcessRunning = () => (
    this.pythonShell &&
    !this.pythonShell.killed &&
    (this.pythonShell.exitCode !== null)
  )

  // Use send a command to the python worker.
  this.send = command => {
    if (!this.isProcessRunning) return { error: 'Process not running' }

    return this.transaction(command => {
      this.pythonShell.stdin.write(
        // Write the command as a JSON object, end with a newline!
        JSON.stringify(command) + '\n'
      )
    }, command)
  }

  this.spawn = async () => {
    console.log('spawning?')
    if (this.isProcessRunning()) return { error: 'Process already running' }
    console.log('spawning!')

    return this.transaction(() => {
      this.pythonShell = spawn(
        `${process.env.PWD}/${process.env.PYTHON_PATH}`,
        [this.workerScript]
      )
      this.pythonShell.stdout.on('data', message => {
        // The python worker returns JSON {data, error}
        console.log('[DATA]', JSON.parse(message))
        this.data.push(JSON.parse(message))
        this.pythonLock.release()
      })
      this.pythonShell.stderr.on('data', error => {
        console.log('[APPERROR]', error)
        this.error.push(serializeError(new Error(error)))
        this.pythonLock.release()
      })
      this.pythonShell.on('close', error => {
        console.log('[CLOSE]', error)
        this.error.push(serializeError(new Error(error)))
        this.pythonLock.release()
      })
      this.pythonShell.on('error', error => {
        console.log('[ERROR]', error)
        this.error.push(serializeError(error))
        this.pythonLock.release()
      })
      console.log('spawned.')
    })
  }

  this.end = async () => {
    if (this.pythonShell.killed) return { error: `process ${this.pythonShell.pid} already killed.` }
    if (this.pythonShell.exitCode !== null) return { error: `process ${this.pythonShell.pid} already exited with code ${this.pythonShell.exitCode}` }
    await this.commandLock.acquire()
    await this.pythonLock.acquire()
    this.pythonShell.stdin.end()
    await this.pythonLock.acquire()
    const pythonError = this.error.pop()
    if (pythonError) {
      this.pythonLock.release()
      this.commandLock.release()
      return { pythonError }
    }
    const { data, error } = this.data.pop()
    this.pythonLock.release()
    this.commandLock.release()
    return { data, error }
  }

  this.kill = async () => {
    if (this.pythonShell.killed) return { error: `process ${this.pythonShell.pid} already killed.` }
    if (this.pythonShell.exitCode !== null) return { error: `process ${this.pythonShell.pid} already exited with code ${this.pythonShell.exitCode}` }
    await this.commandLock.acquire()
    await this.pythonLock.acquire()
    const pythonError = this.error.pop()
    if (pythonError) {
      this.pythonLock.release()
      this.commandLock.release()
      return { pythonError }
    }
    const { data, error } = this.data.pop()
    this.pythonLock.release()
    this.commandLock.release()
    return { data, error }
  }
}

module.exports = PythonWorker
